import os
import csv
import json
import logging
from openpyxl import load_workbook
import pandas as pd
import numpy as np
from typing import Dict

DEFAULT_ENCODING = "utf-8"
DEFAULT_VERBOSE = True
DEFAULT_EXCEL_SHEET_NAME = "Sheet1"
DEFAULT_OVERWRITE = True


def get_field_names(data: list[dict], sort_keys: bool = False):
    key_dict = {}
    for row in data:
        for key in row.keys():
            key_dict[key] = True

    return sorted(list(key_dict.keys())) if sort_keys else list(key_dict.keys())


def write_json(
    file_path: str, data: dict, verbose=DEFAULT_VERBOSE, encoding=DEFAULT_ENCODING
):
    with open(file_path, "w", encoding=encoding) as jsonFile:
        json.dump(data, jsonFile, ensure_ascii=False, indent=2)

        if verbose:
            logging.info(f"Wrote JSON data to {file_path}")


def write_html(file_path, data, verbose=DEFAULT_VERBOSE, encoding=DEFAULT_ENCODING):
    with open(file_path, "w", encoding=encoding) as file:
        file.writelines(data)

    if verbose:
        logging.info(f"Wrote HTML data to {file_path}")


def write_text(file_path, data, verbose=DEFAULT_VERBOSE, encoding=DEFAULT_ENCODING):
    with open(file_path, "w", encoding=encoding) as file:
        file.writelines(data)

    if verbose:
        logging.info(f"Wrote TXT data to {file_path}")


def write_csv(
    file_path: str,
    data: list[dict],
    fieldnames=None,
    overwrite=DEFAULT_OVERWRITE,
    verbose=DEFAULT_VERBOSE,
    encoding=DEFAULT_ENCODING,
):
    if len(data) == 0:
        if verbose:
            logging.info("No CSV data to write")
        return

    write_mode = "w" if overwrite or not os.path.isfile(file_path) else "a"

    with open(file_path, write_mode, newline="", encoding=encoding) as csvfile:
        writer = csv.DictWriter(
            csvfile, fieldnames=fieldnames if fieldnames else get_field_names(data)
        )

        if write_mode == "w":
            writer.writeheader()

        for row in data:
            writer.writerow(row)

        if verbose:
            logging.info(
                f"{'Wrote' if write_mode == 'w' else 'Appended'} CSV data to {file_path} with {len(data)} rows"
            )


def write_excel(
    file_path: str,
    data: list[dict],
    sheet_name=DEFAULT_EXCEL_SHEET_NAME,
    overwrite=DEFAULT_OVERWRITE,
    enable_str_conversion=False,
    verbose=DEFAULT_VERBOSE,
):
    """
    Append data to an existing Excel file or create a new sheet if it doesn't exist.

    Parameters:
    - file_path (str): Path to the existing Excel file.
    - sheet_name (str): Name of the sheet to append data to.
    - data (list of dicts or DataFrame): Data to be appended.
    """

    if len(data) == 0 and not overwrite:
        if verbose:
            logging.info("No EXCEL data to write")
        return

    df = pd.DataFrame(data)

    if enable_str_conversion:
        # Reference: https://stackoverflow.com/a/48454986
        df = df.applymap(
            lambda x: (
                x.encode("unicode_escape").decode("utf-8") if isinstance(x, str) else x
            )
        )

    # If the file does not exist or need to be overwritten, create or overwrite it
    if not os.path.exists(file_path) or overwrite:
        df.to_excel(file_path, sheet_name=sheet_name, index=False)

        if verbose:
            logging.info(f"Wrote EXCEL data to {file_path} with {len(data)} rows")

        return

    # Load the workbook
    book = load_workbook(file_path)

    # Load and append the existing data
    writer = pd.ExcelWriter(file_path, engine="openpyxl")
    writer.book = book

    # If the sheet name already exists, append the data
    if sheet_name in book.sheetnames:
        # Avoid overwriting existing content
        start_row = book[sheet_name].max_row
        # Avoid writing the header if the sheet isn't empty
        header = False if start_row > 0 else True
    # Otherwise, write the data from the first row with header
    else:
        start_row = 0
        header = True

    # Write the DataFrame to the Excel file
    df.to_excel(
        writer, sheet_name=sheet_name, startrow=start_row, header=header, index=False
    )

    # Save the changes
    writer.save()

    if verbose:
        logging.info(f"Wrote EXCEL data to {file_path} with {len(data)} rows")


def write_npy(
    file_path: str,
    data: np.ndarray,
    verbose: bool = DEFAULT_VERBOSE,
    allow_pickle: bool = False,
) -> bool:
    """
    Write a NumPy array to a .npy file.

    Args:
        file_path: Path where the .npy file will be saved.
        data: The NumPy array to save.
        verbose: Whether to log information about the operation.
        allow_pickle: Whether to allow saving object arrays using Python pickles.

    Returns:
        True if the operation was successful, False otherwise.
    """
    try:
        # Create directory if it doesn't exist
        os.makedirs(os.path.dirname(os.path.abspath(file_path)), exist_ok=True)

        np.save(
            file_path,
            data,
            allow_pickle=allow_pickle,
        )

        if verbose:
            logging.info(f"Wrote NumPy array to {file_path} with shape {data.shape}")
        return True
    except Exception as e:
        if verbose:
            logging.error(f"Error writing NumPy file {file_path}: {str(e)}")
        return False


def write_npz(
    file_path: str,
    data: Dict[str, np.ndarray],
    verbose: bool = DEFAULT_VERBOSE,
    allow_pickle: bool = False,
) -> bool:
    """
    Write multiple NumPy arrays to a .npz file.

    Args:
        file_path: Path where the .npz file will be saved.
        data: Dictionary of arrays to save.
        verbose: Whether to log information about the operation.
        allow_pickle: Whether to allow saving object arrays using Python pickles.

    Returns:
        True if the operation was successful, False otherwise.
    """
    try:
        # Create directory if it doesn't exist
        os.makedirs(os.path.dirname(os.path.abspath(file_path)), exist_ok=True)

        np.savez(
            file_path,
            **data,
            allow_pickle=allow_pickle,
        )

        if verbose:
            logging.info(
                f"Wrote NumPy arrays to {file_path} with keys: {list(data.keys())}"
            )
        return True
    except Exception as e:
        if verbose:
            logging.error(f"Error writing NumPy file {file_path}: {str(e)}")
        return False



"""
rexpand-pyutils-file - Python utilities for file system operations.

This package provides a comprehensive set of utilities for file system operations,
including reading, writing, finding, and downloading files.
"""

from rexpand_pyutils_file.file_system import (
    FileSystem,
    read_file,
    write_file,
    download_file,
    unzip_file,
    find_files,
    find_folders,
    refine_path,
)

__version__ = "0.0.7"

__all__ = [
    "FileSystem",
    "read_file",
    "write_file",
    "download_file",
    "unzip_file",
    "find_files",
    "find_folders",
    "refine_path",
]



import pathlib
import platform
from os import listdir, path
from posixpath import join

IGNORED_FILE_SET = set([".DS_Store"])


def refine_path(path: str):
    return str(
        pathlib.WindowsPath(path)
        if platform.system() == "Windows"
        else pathlib.PurePath(path)
    )


def get_file_paths(
    folder_path: str, ignored_file_set=IGNORED_FILE_SET, recursive=False
):
    if not path.exists(folder_path):
        return []

    if not recursive:
        return sorted(
            [
                join(folder_path, file_name)
                for file_name in listdir(folder_path)
                if path.isfile(join(folder_path, file_name))
                and file_name not in ignored_file_set
            ],
        )
    else:

        def internal_get_file_paths(folder_path: str, ignored_file_set=set()):
            file_paths = []
            for item in listdir(folder_path):
                full_path = path.join(folder_path, item)
                if path.isfile(full_path) and item not in ignored_file_set:
                    # Add the file path if it's a file and not ignored
                    file_paths.append(full_path)
                elif path.isdir(full_path):
                    # Recursively search in subdirectories for files
                    file_paths.extend(
                        internal_get_file_paths(full_path, ignored_file_set)
                    )
            return file_paths

        # Get all file paths
        file_paths = internal_get_file_paths(folder_path, ignored_file_set)

        # Sort by the number of path separators (indicating nesting depth)
        # More separators = more nested = higher priority
        # Use character ranking as secondary sorting key
        return sorted(
            file_paths,
            key=lambda x: (
                -x.count(path.sep),
                x,
            ),  # Negative count for reverse order, then lexicographical sort
        )


def get_folder_paths(folder_path: str, ignored_folder_set=set(), recursive=False):
    if not path.exists(folder_path):
        return []

    if not recursive:
        return [
            join(folder_path, folder_name)
            for folder_name in listdir(folder_path)
            if path.isdir(join(folder_path, folder_name))
            and folder_name not in ignored_folder_set
        ]
    else:

        def internal_get_folder_paths(folder_path: str, ignored_folder_set=set()):
            folder_paths = []
            for folder_name in listdir(folder_path):
                full_path = path.join(folder_path, folder_name)
                if path.isdir(full_path) and folder_name not in ignored_folder_set:
                    # Add the current folder path
                    folder_paths.append(full_path)
                    # Recursively add paths from subfolders
                    folder_paths.extend(
                        internal_get_folder_paths(full_path, ignored_folder_set)
                    )
            return folder_paths

        return internal_get_folder_paths(folder_path, ignored_folder_set)



import os
import csv
import json
import logging
import pandas as pd
import numpy as np
from typing import Optional, Dict

DEFAULT_ENCODING = "utf-8"
DEFAULT_VERBOSE = True

# Can be 'ignore', 'replace', or 'backslashreplace'
DEFAULT_READING_ERROR_HANDLING = "backslashreplace"


def read_json(
    file_path: str,
    verbose=DEFAULT_VERBOSE,
    encoding=DEFAULT_ENCODING,
    json_load_extra_args={},
):
    data = None

    if os.path.isfile(file_path):
        with open(
            file_path, encoding=encoding, errors=DEFAULT_READING_ERROR_HANDLING
        ) as jsonFile:
            data = json.load(jsonFile, **json_load_extra_args)

            if verbose:
                logging.info(f"Read JSON data from {file_path}")
    else:
        if verbose:
            logging.warning(f"File not found: {file_path}")

    return data


def read_html(file_path: str, verbose=DEFAULT_VERBOSE, encoding=DEFAULT_ENCODING):
    data = None

    if os.path.isfile(file_path):
        with open(
            file_path, encoding=encoding, errors=DEFAULT_READING_ERROR_HANDLING
        ) as file:
            data = "".join(file.readlines())

        if verbose:
            logging.info(f"Read HTML data from {file_path}")
    else:
        if verbose:
            logging.warning(f"File not found: {file_path}")

    return data


def read_text(file_path: str, verbose=DEFAULT_VERBOSE, encoding=DEFAULT_ENCODING):
    data = None

    if os.path.isfile(file_path):
        with open(
            file_path, encoding=encoding, errors=DEFAULT_READING_ERROR_HANDLING
        ) as file:
            data = "".join(file.readlines())

        if verbose:
            logging.info(f"Read TXT data from {file_path}")
    else:
        if verbose:
            logging.warning(f"File not found: {file_path}")

    return data


def read_csv(
    file_path: str,
    delimiter: str = ",",
    quotechar: str = '"',
    escapechar: str | None = None,
    doublequote: bool = True,
    skipinitialspace: bool = False,
    lineterminator: str = "\r\n",
    quoting=0,
    strict: bool = False,
    verbose=DEFAULT_VERBOSE,
    encoding=DEFAULT_ENCODING,
    filter_na: bool = False,
    replace_na: bool = True,
    strip_bom: bool = True,
) -> list[dict] | None:
    data = None

    if os.path.isfile(file_path):
        with open(
            file_path,
            newline="",
            encoding=encoding,
            errors=DEFAULT_READING_ERROR_HANDLING,
        ) as csvfile:
            spamreader = csv.DictReader(
                csvfile,
                delimiter=delimiter,
                quotechar=quotechar,
                escapechar=escapechar,
                doublequote=doublequote,
                skipinitialspace=skipinitialspace,
                lineterminator=lineterminator,
                quoting=quoting,
                strict=strict,
            )

            # Handle BOM character in column names if needed
            if strip_bom:
                # Get the fieldnames and strip BOM from them
                fieldnames = spamreader.fieldnames
                if fieldnames and any(name.startswith("\ufeff") for name in fieldnames):
                    # Create a mapping of original names to cleaned names
                    name_mapping = {
                        name: name.replace("\ufeff", "") for name in fieldnames
                    }
                    # Create a new DictReader with cleaned fieldnames
                    spamreader = csv.DictReader(
                        csvfile,
                        fieldnames=[name_mapping[name] for name in fieldnames],
                        delimiter=delimiter,
                        quotechar=quotechar,
                        escapechar=escapechar,
                        doublequote=doublequote,
                        skipinitialspace=skipinitialspace,
                        lineterminator=lineterminator,
                        quoting=quoting,
                        strict=strict,
                    )
                    # Reset file pointer to beginning
                    csvfile.seek(0)
                    # Skip header row
                    next(spamreader)

            data = [row for row in spamreader]

            if filter_na:
                # Filter out rows with NA values
                filtered_data = []
                for row in data:
                    # Check if any value in the row is None, empty string, or 'NA'
                    if not any(
                        val is None or val == "" or val.upper() == "NA"
                        for val in row.values()
                    ):
                        filtered_data.append(row)
                data = filtered_data
            elif replace_na:
                # Replace NA values with empty strings
                for row in data:
                    for key in row:
                        if (
                            row[key] is None
                            or row[key] == ""
                            or str(row[key]).upper() == "NA"
                        ):
                            row[key] = ""

            if verbose:
                logging.info(f"Read CSV data from {file_path} with {len(data)} rows")
    else:
        if verbose:
            logging.warning(f"File not found: {file_path}")

    return data


def read_excel(
    file_path: str,
    sheet_name: str | None = None,
    filter_na: bool = False,
    replace_na: bool = True,
    verbose=DEFAULT_VERBOSE,
) -> list[dict] | None:
    data = None

    if os.path.isfile(file_path):
        # If sheet_name is None, read the first sheet
        if sheet_name is None:
            # Read the Excel file without specifying a sheet
            df = pd.read_excel(file_path, sheet_name=0)
            if verbose:
                logging.info(f"Read first sheet from Excel file {file_path}")
        else:
            # Read the specified sheet
            df = pd.read_excel(file_path, sheet_name=sheet_name)
            if verbose:
                logging.info(f"Read sheet '{sheet_name}' from Excel file {file_path}")

        if filter_na:
            # Drop rows with any NA values
            df = df.dropna()
        elif replace_na:
            # Replace NA values with empty strings
            df = df.fillna("")

        data = df.to_dict(
            orient="records",
        )

        if verbose:
            logging.info(f"Read EXCEL data from {file_path} with {len(data)} rows")
    else:
        if verbose:
            logging.warning(f"File not found: {file_path}")

    return data


def read_npy(
    file_path: str,
    verbose: bool = DEFAULT_VERBOSE,
    allow_pickle: bool = False,
    encoding: str = "ASCII",
) -> Optional[np.ndarray]:
    """
    Read a NumPy array from a .npy file.

    Args:
        file_path: Path to the .npy file.
        verbose: Whether to log information about the operation.
        allow_pickle: Whether to allow loading pickled object arrays.
        encoding: What encoding to use when reading Python 2 strings.

    Returns:
        The loaded NumPy array, or None if the file doesn't exist.
    """
    if not os.path.isfile(file_path):
        if verbose:
            logging.warning(f"File not found: {file_path}")
        return None

    try:
        data = np.load(
            file_path,
            allow_pickle=allow_pickle,
            encoding=encoding,
        )
        if verbose:
            logging.info(f"Read NumPy array from {file_path} with shape {data.shape}")
        return data
    except Exception as e:
        if verbose:
            logging.error(f"Error reading NumPy file {file_path}: {str(e)}")
        return None


def read_npz(
    file_path: str,
    verbose: bool = DEFAULT_VERBOSE,
    allow_pickle: bool = False,
    encoding: str = "ASCII",
) -> Optional[Dict[str, np.ndarray]]:
    """
    Read multiple NumPy arrays from a .npz file.

    Args:
        file_path: Path to the .npz file.
        verbose: Whether to log information about the operation.
        allow_pickle: Whether to allow loading pickled object arrays.
        encoding: What encoding to use when reading Python 2 strings.

    Returns:
        A dictionary of arrays, or None if the file doesn't exist.
    """
    if not os.path.isfile(file_path):
        if verbose:
            logging.warning(f"File not found: {file_path}")
        return None

    try:
        data = np.load(
            file_path,
            allow_pickle=allow_pickle,
            encoding=encoding,
        )

        # Convert to dictionary
        result = {key: data[key] for key in data.files}

        if verbose:
            logging.info(
                f"Read NumPy arrays from {file_path} with keys: {list(result.keys())}"
            )
        return result
    except Exception as e:
        if verbose:
            logging.error(f"Error reading NumPy file {file_path}: {str(e)}")
        return None



import mimetypes
import logging
import requests

from .utils import get_or_create_folder_path

DEFAULT_VERBOSE = True


def download_file(url, file_path_without_extension, verbose=DEFAULT_VERBOSE):
    # Send a GET request
    response = requests.get(url, stream=True)
    response.raise_for_status()  # Ensure the request was successful

    # Try to extract the file extension from the Content-Type header
    content_type = response.headers.get("Content-Type")
    extension = mimetypes.guess_extension(content_type) if content_type else None

    # If we couldn't find an extension, default to '.bin'
    if not extension:
        extension = ".bin"

    if verbose:
        logging.info(f"Guessed the file extension as {extension}")

    get_or_create_folder_path(file_path_without_extension)

    file_path = f"{file_path_without_extension}{extension}"

    if verbose:
        logging.info(f"Downloading file from {url}")

    # Write the response content to a file
    with open(file_path, "wb") as file:
        for chunk in response.iter_content(chunk_size=8192):
            file.write(chunk)

        if verbose:
            logging.info(f"Downloaded file to {file_path}")

    return file_path



"""
FileSystem module for rexpand-pyutils-file.

This module provides the FileSystem class for file system operations.
"""

import gzip
import logging
from os.path import isfile

from rexpand_pyutils_file.utils import get_or_create_folder_path
from rexpand_pyutils_file.downloader import download_file as static_download_file
from rexpand_pyutils_file.finder import (
    IGNORED_FILE_SET,
    get_file_paths,
    get_folder_paths,
    refine_path as static_refine_path,
)
from rexpand_pyutils_file.reader import (
    DEFAULT_ENCODING,
    DEFAULT_VERBOSE,
    read_text,
    read_csv,
    read_excel,
    read_html,
    read_json,
    read_npy,
    read_npz,
)
from rexpand_pyutils_file.writer import (
    DEFAULT_OVERWRITE,
    DEFAULT_EXCEL_SHEET_NAME,
    write_csv,
    write_excel,
    write_html,
    write_json,
    write_text,
    write_npy,
    write_npz,
)


class FileSystem:
    """
    A class for file system operations.

    This class provides methods for reading, writing, finding, and downloading files.
    It also provides methods for unzipping files and refining paths.
    """

    def __init__(self, path=None, verbose=DEFAULT_VERBOSE, encoding=DEFAULT_ENCODING):
        """
        Initialize the FileSystem class.

        Args:
            path (str, optional): The default path to use for operations. Defaults to None.
            verbose (bool, optional): Whether to print verbose output. Defaults to DEFAULT_VERBOSE.
            encoding (str, optional): The encoding to use for file operations. Defaults to DEFAULT_ENCODING.
        """
        self.path = path
        self.verbose = verbose
        self.encoding = encoding

    def __get_path(self, path):
        """
        Get the path to use for operations.

        Args:
            path (str, optional): The path to use. If None, use the default path.

        Returns:
            str: The path to use for operations.

        Raises:
            Exception: If no path is specified.
        """
        path = path if path is not None else self.path

        if not path:
            raise Exception("Path must be specified")

        return path

    def read(
        self,
        path=None,
        verbose=None,
        encoding=None,
        excel_sheet_name=None,
    ):
        """
        Read a file.

        Args:
            path (str, optional): The path to the file. If None, use the default path.
            verbose (bool, optional): Whether to print verbose output. If None, use the default verbose setting.
            encoding (str, optional): The encoding to use. If None, use the default encoding.
            excel_sheet_name (str, optional): The name of the sheet to read from an Excel file. If None, read the first sheet.

        Returns:
            The contents of the file.

        Raises:
            Exception: If the file type is not supported.
        """
        path = self.__get_path(path)
        verbose = verbose if verbose is not None else self.verbose
        encoding = encoding if encoding is not None else self.encoding

        if isfile(path):
            if path[-5:] == ".json":
                return read_json(file_path=path, verbose=verbose, encoding=encoding)
            elif path[-5:] == ".html":
                return read_html(file_path=path, verbose=verbose, encoding=encoding)
            elif path[-4:] == ".txt" or path[-4:] == ".log" or path[-3:] == ".md" or path[-4:] == ".mmd":
                return read_text(file_path=path, verbose=verbose, encoding=encoding)
            elif path[-4:] == ".csv":
                return read_csv(file_path=path, verbose=verbose, encoding=encoding)
            elif path[-5:] == ".xlsx":
                return read_excel(
                    file_path=path, sheet_name=excel_sheet_name, verbose=verbose
                )
            elif path[-4:] == ".npy":
                return read_npy(file_path=path, verbose=verbose)
            elif path[-4:] == ".npz":
                return read_npz(file_path=path, verbose=verbose)
            else:
                raise Exception(f"Unsupported file type for {path}")
        else:
            if verbose:
                logging.warning(f"File not found: {path}")
            return None

    def write(
        self,
        data,
        path=None,
        verbose=None,
        csv_and_excel_overwrite=DEFAULT_OVERWRITE,
        encoding=None,
        csv_fieldnames=None,
        excel_sheet_name=DEFAULT_EXCEL_SHEET_NAME,
        excel_enable_str_conversion=False,
    ):
        """
        Write data to a file.

        Args:
            data: The data to write.
            path (str, optional): The path to the file. If None, use the default path.
            verbose (bool, optional): Whether to print verbose output. If None, use the default verbose setting.
            csv_and_excel_overwrite (bool, optional): Whether to overwrite existing CSV or Excel files. Defaults to DEFAULT_OVERWRITE.
            encoding (str, optional): The encoding to use. If None, use the default encoding.
            csv_fieldnames (list, optional): The fieldnames to use for CSV files. If None, use the keys from the data.
            excel_sheet_name (str, optional): The name of the sheet to write to an Excel file. Defaults to DEFAULT_EXCEL_SHEET_NAME.
            excel_enable_str_conversion (bool, optional): Whether to enable string conversion for Excel files. Defaults to False.

        Returns:
            None

        Raises:
            Exception: If the file type is not supported.
        """
        path = self.__get_path(path)
        verbose = verbose if verbose is not None else self.verbose
        encoding = encoding if encoding is not None else self.encoding

        get_or_create_folder_path(path)

        if path[-5:] == ".json":
            return write_json(
                file_path=path, data=data, verbose=verbose, encoding=encoding
            )
        elif path[-5:] == ".html":
            return write_html(
                file_path=path, data=data, verbose=verbose, encoding=encoding
            )
        elif path[-4:] == ".txt" or path[-4:] == ".log" or path[-3:] == ".md" or path[-4:] == ".mmd":
            return write_text(
                file_path=path, data=data, verbose=verbose, encoding=encoding
            )
        elif path[-4:] == ".csv":
            return write_csv(
                file_path=path,
                data=data,
                fieldnames=csv_fieldnames,
                verbose=verbose,
                overwrite=csv_and_excel_overwrite,
                encoding=encoding,
            )
        elif path[-5:] == ".xlsx":
            return write_excel(
                file_path=path,
                data=data,
                sheet_name=excel_sheet_name,
                verbose=verbose,
                overwrite=csv_and_excel_overwrite,
                enable_str_conversion=excel_enable_str_conversion,
            )
        elif path[-4:] == ".npy":
            return write_npy(file_path=path, data=data, verbose=verbose)
        elif path[-4:] == ".npz":
            return write_npz(file_path=path, data=data, verbose=verbose)
        else:
            raise Exception(f"Unsupported file type for {path}")

    def unzip(self, path=None, output_path=None, verbose=None):
        """
        Unzip a gzip file.

        Args:
            path (str, optional): The path to the gzip file. If None, use the default path.
            output_path (str, optional): The path to write the unzipped data to. If None, return the unzipped data.
            verbose (bool, optional): Whether to print verbose output. If None, use the default verbose setting.

        Returns:
            The unzipped data if output_path is None, otherwise None.

        Raises:
            Exception: If the file type is not supported.
        """
        path = self.__get_path(path)
        verbose = verbose if verbose is not None else self.verbose

        if not isfile(path):
            if verbose:
                logging.warning(f"File not found: {path}")
            return None

        get_or_create_folder_path(path)

        if path[-3:] == ".gz":
            with gzip.open(path, "rb") as gz_file:
                data = gz_file.read()

                if output_path:
                    with open(output_path, "wb") as out_file:
                        out_file.write(data)

                    if verbose:
                        logging.info(
                            f"Unzipped data from {path} and wrote data to {output_path}"
                        )
                else:
                    if verbose:
                        logging.info(f"Unzipped data from {path} ")

                return data
        else:
            raise Exception(f"Unsupported file type for {path}")

    def download(self, url, file_path_without_extension=None, verbose=None):
        """
        Download a file from a URL.

        Args:
            url (str): The URL to download from.
            file_path_without_extension (str, optional): The path to save the file to, without the extension. If None, use the default path.
            verbose (bool, optional): Whether to print verbose output. If None, use the default verbose setting.

        Returns:
            The path to the downloaded file.
        """
        path = self.__get_path(file_path_without_extension)
        verbose = verbose if verbose is not None else self.verbose

        get_or_create_folder_path(path)

        return static_download_file(
            url=url, file_path_without_extension=path, verbose=verbose
        )

    def find_files(
        self,
        path=None,
        ignored_file_set=IGNORED_FILE_SET,
        file_suffix=None,
        recursive=False,
    ):
        """
        Find files in a directory.

        Args:
            path (str, optional): The path to the directory. If None, use the default path.
            ignored_file_set (set, optional): A set of file names to ignore. Defaults to IGNORED_FILE_SET.
            file_suffix (str, optional): A suffix to filter files by. If None, return all files.
            recursive (bool, optional): Whether to search recursively. Defaults to False.

        Returns:
            A list of file paths.
        """
        path = path if path is not None else self.path
        file_paths = get_file_paths(path, ignored_file_set, recursive)
        if file_suffix:
            if file_suffix[0] != ".":
                file_suffix = "." + file_suffix

            return [
                file_path
                for file_path in file_paths
                if file_path[-len(file_suffix) :] == file_suffix
            ]

        return file_paths

    def find_folders(self, path=None, ignored_folder_set=set(), recursive=False):
        """
        Find folders in a directory.

        Args:
            path (str, optional): The path to the directory. If None, use the default path.
            ignored_folder_set (set, optional): A set of folder names to ignore. Defaults to an empty set.
            recursive (bool, optional): Whether to search recursively. Defaults to False.

        Returns:
            A list of folder paths.
        """
        path = path if path is not None else self.path
        return get_folder_paths(path, ignored_folder_set, recursive)

    def refine_path(self, path=None):
        """
        Refine a path.

        Args:
            path (str, optional): The path to refine. If None, use the default path.

        Returns:
            The refined path.
        """
        path = path if path is not None else self.path
        return static_refine_path(path)


# Convenience functions that use the FileSystem class
def read_file(path, verbose=None, encoding=None, excel_sheet_name=None):
    """
    Read a file.

    Args:
        path (str): The path to the file.
        verbose (bool, optional): Whether to print verbose output. Defaults to None.
        encoding (str, optional): The encoding to use. Defaults to None.
        excel_sheet_name (str, optional): The name of the sheet to read from an Excel file. If None, read the first sheet.

    Returns:
        The contents of the file.
    """
    return FileSystem().read(
        path=path, verbose=verbose, encoding=encoding, excel_sheet_name=excel_sheet_name
    )


def write_file(
    path,
    data,
    verbose=None,
    encoding=None,
    csv_and_excel_overwrite=DEFAULT_OVERWRITE,
    csv_fieldnames=None,
    excel_sheet_name=DEFAULT_EXCEL_SHEET_NAME,
    excel_enable_str_conversion=False,
):
    """
    Write data to a file.

    Args:
        path (str): The path to the file.
        data: The data to write.
        verbose (bool, optional): Whether to print verbose output. Defaults to None.
        encoding (str, optional): The encoding to use. Defaults to None.
        csv_and_excel_overwrite (bool, optional): Whether to overwrite existing CSV or Excel files. Defaults to DEFAULT_OVERWRITE.
        csv_fieldnames (list, optional): The fieldnames to use for CSV files. If None, use the keys from the data.
        excel_sheet_name (str, optional): The name of the sheet to write to an Excel file. Defaults to DEFAULT_EXCEL_SHEET_NAME.
        excel_enable_str_conversion (bool, optional): Whether to enable string conversion for Excel files. Defaults to False.

    Returns:
        None
    """
    return FileSystem().write(
        data=data,
        path=path,
        verbose=verbose,
        csv_and_excel_overwrite=csv_and_excel_overwrite,
        encoding=encoding,
        csv_fieldnames=csv_fieldnames,
        excel_sheet_name=excel_sheet_name,
        excel_enable_str_conversion=excel_enable_str_conversion,
    )


def download_file(url, file_path_without_extension, verbose=None):
    """
    Download a file from a URL.

    Args:
        url (str): The URL to download from.
        file_path_without_extension (str): The path to save the file to, without the extension.
        verbose (bool, optional): Whether to print verbose output. Defaults to None.

    Returns:
        The path to the downloaded file.
    """
    return FileSystem().download(
        url=url,
        file_path_without_extension=file_path_without_extension,
        verbose=verbose,
    )


def unzip_file(path, output_path=None, verbose=None):
    """
    Unzip a gzip file.

    Args:
        path (str): The path to the gzip file.
        output_path (str, optional): The path to write the unzipped data to. If None, return the unzipped data.
        verbose (bool, optional): Whether to print verbose output. Defaults to None.

    Returns:
        The unzipped data if output_path is None, otherwise None.
    """
    return FileSystem().unzip(path=path, output_path=output_path, verbose=verbose)


def find_files(
    path, ignored_file_set=IGNORED_FILE_SET, file_suffix=None, recursive=False
):
    """
    Find files in a directory.

    Args:
        path (str): The path to the directory.
        ignored_file_set (set, optional): A set of file names to ignore. Defaults to IGNORED_FILE_SET.
        file_suffix (str, optional): A suffix to filter files by. If None, return all files.
        recursive (bool, optional): Whether to search recursively. Defaults to False.

    Returns:
        A list of file paths.
    """
    return FileSystem().find_files(
        path=path,
        ignored_file_set=ignored_file_set,
        file_suffix=file_suffix,
        recursive=recursive,
    )


def find_folders(path, ignored_folder_set=set(), recursive=False):
    """
    Find folders in a directory.

    Args:
        path (str): The path to the directory.
        ignored_folder_set (set, optional): A set of folder names to ignore. Defaults to an empty set.
        recursive (bool, optional): Whether to search recursively. Defaults to False.

    Returns:
        A list of folder paths.
    """
    return FileSystem().find_folders(
        path=path, ignored_folder_set=ignored_folder_set, recursive=recursive
    )


def refine_path(path):
    """
    Refine a path.

    Args:
        path (str): The path to refine.

    Returns:
        The refined path.
    """
    return FileSystem().refine_path(path=path)
